import xml.etree.ElementTree as ET
from pathlib import Path


class Trainer:
    def __init__(
            self,
            model_script_path: str = "",
            model_option_path: str = "",
            model_option_string: str = "",
            model_weights_path: str = "",
            model_stream: int = 0,
            model_create: str = "PythonModel",
            model_multirole_input = False,
            users: list = None,
            classes: list = None,
            streams: list = None,
            register: list = None,
            info_trained: bool = False,
            meta_right_ctx: int = 0,
            meta_left_ctx: int = 0,
            meta_balance: str = "none",
            meta_backend: str = "nova-server",
            ssi_v="5",
            xml_version="1.0",
    ):

        self.model_multi_role_input = None
        self.model_script_path = model_script_path
        self.model_option_path = model_option_path
        self.model_optstr = model_option_string
        self.model_weights_path = model_weights_path
        self.model_stream = model_stream
        self.model_create = model_create
        self.users = users if users is not None else []
        self.classes = classes if classes is not None else []
        self.streams = streams if streams is not None else []
        self.register = register if register is not None else []
        self.info_trained = info_trained
        self.meta_right_ctx = meta_right_ctx
        self.meta_left_ctx = meta_left_ctx
        self.meta_balance = meta_balance
        self.meta_backend = meta_backend
        self.ssi_v = ssi_v
        self.xml_version = xml_version
        self.model_multi_role_input = model_multirole_input

    def load_from_file(self, fp):
        root = ET.parse(Path(fp))
        info = root.find("info")
        meta = root.find("meta")
        register = root.find("register")
        streams = root.find("streams")
        classes = root.find("classes")
        users = root.find("users")
        model = root.find("model")

        if info is not None:
            self.info_trained = info.get("trained")
        if meta is not None:
            self.meta_left_ctx = int(meta.get("leftContext", default="0"))
            self.meta_right_ctx = int(meta.get("rightContext", default="0"))
            self.meta_balance = meta.get("balance", default="none")
            self.meta_backend = meta.get("backend", default="Python")
        if register is not None:
            for r in register:
                self.register.append(r.attrib)
        if streams is not None:
            for s in streams:
                self.streams.append(s.attrib)
        if classes is not None:
            for c in classes:
                self.classes.append(c.attrib)
        if users is not None:
            for u in users:
                self.users.append(u.attrib)
        if model is not None:
            self.model_stream = model.get("stream", default="0")
            self.model_create = model.get("create", default="PythonModel")
            self.model_option_path = model.get("option", default="")
            self.model_script_path = model.get("script", default="")
            self.model_weights_path = model.get("path", default="")
            self.model_optstr = model.get("optstr", default="")
            self.model_multi_role_input = bool(model.get("multi_role_input", default=False))

    def write_to_file(self, fp):
        root = ET.Element("trainer")
        ET.SubElement(root, "info", trained=str(self.info_trained))
        ET.SubElement(
            root,
            "meta",
            leftContext=str(self.meta_left_ctx),
            rightContex=str(self.meta_right_ctx),
            balance=self.meta_balance,
            backend=self.meta_backend,
        )
        register = ET.SubElement(root, "register")
        for r in self.register:
            ET.SubElement(register, "item", **r)
        streams = ET.SubElement(root, "streams")
        for s in self.streams:
            ET.SubElement(streams, "item", **s)
        classes = ET.SubElement(root, "classes")
        for c in self.classes:
            ET.SubElement(classes, "item", **c)
        users = ET.SubElement(root, "users")
        for u in self.users:
            ET.SubElement(users, "item", **u)
        ET.SubElement(
            root,
            "model",
            create=self.model_create,
            stream=str(self.model_stream),
            path=self.model_weights_path,
            script=self.model_script_path,
            optstr=self.model_optstr,
            option=self.model_option_path,
        )

        tree = ET.ElementTree(root)
        ET.indent(tree, space="    ", level=0)

        if not fp.suffix:
            fp = fp.with_suffix(".trainer")
        tree.write(fp)


class ChainLink:

    def __init__(self, create: str, script: str, optsstr: str, syspath: str, tag: str = "feature", multi_role_input: str = 'False', **kwargs):
        self.create = create
        self.script = script
        self.optsstr = optsstr
        self.syspath = syspath
        self.tag = tag
        self.multi_role_input = True if multi_role_input == 'True' else False


class Chain:
    def __init__(
        self,
        meta_frame_step: str = "",
        meta_left_context: str = "",
        meta_right_context: str = "",
        meta_backend: str = "nova-server",
        meta_description: str = "",
        meta_category: str = "",
        register: list = None,
        links: list = None,
    ):
        self.meta_frame_step = meta_frame_step
        self.meta_left_ctx = meta_left_context
        self.meta_right_ctx = meta_right_context
        self.meta_backend = meta_backend
        self.meta_description = meta_description
        self.meta_category = meta_category
        self.register = register if register else []
        self.links = links if links else []

    def load_from_file(self, fp):
        tree = ET.parse(Path(fp))
        root = tree.getroot()
        meta = tree.find("meta")
        register = tree.find("register")
        links = []
        for child in root:
            if child.tag == "feature" or child.tag == "filter":
                links.append(child)

        if meta is not None:
            self.meta_frame_step = meta.attrib.get("frameStep", "0")
            self.meta_left_ctx = meta.attrib.get("leftContext", "0")
            self.meta_right_ctx = meta.attrib.get("rightContext", "0")
            self.meta_backend = meta.attrib.get("backend", "nova-server")
            self.meta_description = meta.attrib.get("description", "")
            self.meta_category = meta.attrib.get("category", "")

        if register is not None:
            for r in register:
                self.register.append(r.attrib)

        for link in links:
            item = link.find("item")
            new_link = ChainLink(**item.attrib, tag=link.tag)
            self.links.append(new_link)

    def write_to_file(self, fp):
        root = ET.Element("chain")
        ET.SubElement(
            root,
            "meta",
            frameStep=str(self.meta_frame_step),
            leftContext=str(self.meta_left_ctx),
            rightContex=str(self.meta_right_ctx),
            backend=str(self.meta_backend),
            description=str(self.meta_description),
            category=str(self.meta_category),

        )
        register = ET.SubElement(root, "register")
        for r in self.register:
            ET.SubElement(register, "item", **r)

        cl: ChainLink
        for cl in self.links:
            link = ET.SubElement(root, cl.tag)
            ET.SubElement(link, "item", create=cl.create, script=cl.script, syspath=cl.syspath, optsstr=cl.optsstr, multi_role_input=str(cl.multi_role_input))

        tree = ET.ElementTree(root)
        ET.indent(tree, space="    ", level=0)

        if not fp.suffix:
            fp = fp.with_suffix(".chain")
        tree.write(fp)


if __name__ == "__main__":
    chain_in_fp = Path("/Users/dominikschiller/Work/github/nova-server/local/cml/chains/test/uc1/uc1.chain")
    chain_out_fp = Path("test_chain.chain")

    chain = Chain()
    chain.load_from_file(chain_in_fp)
    chain.write_to_file(chain_out_fp)
    breakpoint()

    trainer_in_fp = Path(
        r"Z:\nova\cml\models\trainer\discrete\base_emotions\feature{compare[480ms,40ms,480ms]}\linsvm\linsvm.compare.trainer"
    )
    trainer_out_fp = Path(".test_trainer.trainer")

    trainer = Trainer()
    trainer.load_from_file(trainer_in_fp)
    trainer.write_trainer_to_file(trainer_out_fp)



from enum import Enum

class DataTypes(Enum):
  VIDEO = 0
  AUDIO  = 1
  FEATURE = 2


class AnnoTypes(Enum):
  DISCRETE = 0
  CONTINUOUS = 1
  FREE = 2
  POINT = 3
  DISCRETE_POLYGON = 4


'''Helper'''
def string_to_enum(enum, string):
  """

  Searches for 'string' in the enum names and returns the corresponding enum object.
  If the passed string does not correspond to any name in the enum the last item in the enum will be instantiated and returned.

  Args:
    enum (): The enum class to compare against
    string (str): String representation of the enum member

  Returns:
    enum: Instance of the provided enum class
  """
  for e in enum:
    if e.name == string.upper():
      return e

  print(f'Warning! Unknown type {string}. Assuming {e} data type.')
  return e



import xml.etree.ElementTree as ET
import numpy as np
from nova_utils.ssi_utils.ssi_data_types import NPDataTypes, FileTypes, string_to_enum
from pathlib import Path

class Chunk:
    def __init__(self, f=0, t=0, b=0, n=0):
        self.f = f
        self.t = t
        self.b = b
        self.n = n


class Stream:
    def __init__(
        self,
        ftype: FileTypes = FileTypes.UNDEF,
        sr: int = 0,
        dim: int = 0,
        byte: int = 4,
        dtype: NPDataTypes = NPDataTypes.UNDEF,
        delim: str = "",
        chunks: list = None,
        data: np.ndarray = None,
        path: Path =None,
    ):
        self.ftype = ftype
        self.sr = sr
        self.dim = dim
        self.byte = byte
        self.dtype = dtype
        self.delim = delim
        self.chunks = chunks if chunks else []
        self.data = data

        if path:
            self.load(path)

    def load_header(self, path):
        tree = ET.parse(path)
        root = tree.getroot()
        chunks = 0

        for child in root:
            if child.tag == "info":
                for key, val in child.attrib.items():
                    if key == "ftype":
                        self.ftype = string_to_enum(FileTypes, val)
                    elif key == "sr":
                        self.sr = float(val)
                    elif key == "dim":
                        self.dim = int(val)
                    elif key == "byte":
                        self.byte = int(val)
                    elif key == "type":
                        self.dtype = string_to_enum(NPDataTypes, val).value
                    elif key == "delim":
                        self.delim = val
            elif child.tag == "chunk":
                f, t, b, n = 0, 0, 0, 0
                for key, val in child.attrib.items():
                    if key == "from":
                        f = float(val)
                    elif key == "to":
                        t = float(val)
                    elif key == "num":
                        n = int(val)
                    elif key == "byte":
                        b = int(val)
                chunks += 1
                self.chunks.append(Chunk(f, t, b, n))

    def load_data(self, path):
        if self.ftype == FileTypes.ASCII:
            self.data = np.loadtxt(path, dtype=self.dtype, delimiter=self.delim)
        elif self.ftype == FileTypes.BINARY:
            # number of all data chunks
            num = sum([c.n for c in self.chunks])
            self.data = np.fromfile(path, dtype=self.dtype).reshape(num, self.dim)
        else:
            raise ValueError("FileType {} not supported".format(self))

    def load(self, path: Path):
        if not path == type(Path):
            path = Path(path)
        self.load_header(path)
        self.load_data(path.parent / (path.name + '~'))

    def save_header(self, path):

        root = ET.Element("stream")
        ET.SubElement(
            root,
            "info",
            ftype=str(self.ftype.name),
            sr=str(self.sr),
            dim=str(self.dim),
            byte=str(self.byte),
            type=NPDataTypes(self.dtype).name,
            # delim=str(self.delim),
        )
        ET.SubElement(
            root,
            "meta",
        )

        for chunk in self.chunks:
            chunk = {
                "from": str(chunk.f),
                "to": str(chunk.t),
                "byte": str(chunk.b),
                "num": str(chunk.n),
            }

            ET.SubElement(root, "chunk", **chunk)

        tree = ET.ElementTree(root)
        ET.indent(tree, space="    ", level=0)
        tree.write(path)

    def save_data(self, path):
        if self.ftype == FileTypes.ASCII:
            np.savetxt(path, self.data, delimiter=self.delim)
        if self.ftype == FileTypes.BINARY:
            self.data.tofile(path)

    def save(self, path: Path):
        if not path == type(Path):
            path = Path(path)
        if not path.suffix == '.stream':
            path = path.parent / (path.name + '.stream')
        self.save_header(path)
        self.save_data(path.parent / (path.name + '~'))


if __name__ == "__main__":
    stream = Stream()
    stream.load(Path("../../local/novice.audio.gemaps[480ms,40ms,480ms].stream"))

    # modify data
    stream.data = np.random.rand(sum([c.n for c in stream.chunks]), stream.dim).astype(
        np.float32
    )
    stream.ftype = FileTypes.ASCII

    stream.save(Path("../../local/novice.audio.gemaps[480ms,40ms,480ms]_modified.stream"))



import xml.etree.ElementTree as ET
from nova_utils.ssi_utils.ssi_data_types import FileTypes, string_to_enum
from enum import Enum
import numpy as np
import csv
from struct import *


class LabelDataType(Enum):
    DISCRETE = np.dtype(
        [
            ("from", np.float64),
            ("to", np.float64),
            ("id", np.int32),
            ("conf", np.float32),
        ]
    )  # {'names':('from', 'to', 'name', 'conf'),

    FREE = np.dtype(
        [
            ("from", np.float64),
            ("to", np.float64),
            ("name", np.object_),
            ("conf", np.float32),
        ]
    )

    CONTINUOUS = np.dtype([("score", np.float32), ("conf", np.float32)])


class SchemeType(Enum):
    DISCRETE = 0
    CONTINUOUS = 1
    FREE = 2


class Scheme:
    def __init__(
        self,
        name: str = "",
        type: SchemeType = SchemeType.DISCRETE,
        sr: float = 0,
        min: float = 0,
        max: float = 0,
        classes: dict = None,
    ):
        self.name = name
        self.type = type
        self.sr = sr
        self.min = min
        self.max = max
        self.classes = classes if classes else {}
        self.dtype = None

    def get_dtype(self):
        return LabelDataType[self.type.name].value


class Anno:
    ftype: FileTypes

    # def __init__(self, ftype="UNDEF", size=0, role="", annotator=""):
    def __init__(
        self,
        ftype: FileTypes = FileTypes.UNDEF,
        size: int = 0,
        role: str = "",
        annotator: str = "",
        data: list = None,
        scheme: Scheme = None,
    ):
        self.ftype = ftype
        self.size = size
        self.role = role
        self.annotator = annotator
        self.scheme = scheme if scheme else Scheme()
        self.data = None
        if data:
            self.set_data(data)

    def _verify_data_format(self, data: np.array):
        assert len(data) > 0
        if self.scheme.type == SchemeType.DISCRETE:
            data.dtype = LabelDataType.DISCRETE.value
        if self.scheme.type == SchemeType.CONTINUOUS:
            data.dtype = LabelDataType.CONTINUOUS.value
        if self.scheme.type == SchemeType.FREE:
            data.dtype = LabelDataType.FREE.value

        return True

    def set_data(self, data: [np.array, list]):
        """
        Helper function to add data to an existing annotation. This function checks the format of the array and sets additional attributes (e.g. min and max value if necesary).
        Needs to be called after the scheme has been set initialized. If the input is passed as a list the function will try to convert it to an np.array.
        Args:
            data (np.array, list): The annotation data to set
        """

        assert self.scheme is not None and self.scheme.type is not None
        if type(data) is list:
            data = np.asarray(data, self.scheme.get_dtype())

        if self._verify_data_format(data):
            self.data = data

    def load_header(self, path):
        tree = ET.parse(path)
        root = tree.getroot()

        for child in root:
            for key, val in child.attrib.items():
                if child.tag == "info":
                    if key == "ftype":
                        self.ftype = string_to_enum(FileTypes, val)
                    elif key == "size":
                        self.size = int(val)
                elif child.tag == "meta":
                    if key == "role":
                        self.role = val
                    elif key == "annotator":
                        self.annotator = val
                elif child.tag == "scheme":
                    if key == "name":
                        self.scheme.name = val
                    if key == "type":
                        self.scheme.type = string_to_enum(SchemeType, val)
                    if key == "sr":
                        self.scheme.sr = float(val)
                    if key == "min":
                        self.scheme.min = int(val)
                    if key == "max":
                        self.scheme.max = int(val)
            if child.tag == "scheme" and self.scheme.type == SchemeType.DISCRETE:
                for item in child:
                    for key, val in item.attrib.items():
                        if key == "name":
                            class_name = val
                        elif key == "id":
                            id = val
                    self.scheme.classes[id] = class_name

    def load_data_discrete(self, path):
        dt = LabelDataType.DISCRETE.value
        if self.ftype == FileTypes.ASCII:
            self.data = np.loadtxt(path, dtype=dt, delimiter=";")
        elif self.ftype == FileTypes.BINARY:
            self.data = np.fromfile(path, dtype=dt)
        else:
            raise ValueError("FileType {} not supported".format(self.ftype))

    def load_data_continuous(self, path):
        dt = LabelDataType.CONTINUOUS.value
        if self.ftype == FileTypes.ASCII:
            self.data = np.loadtxt(path, dtype=dt, delimiter=";")
        elif self.ftype == FileTypes.BINARY:
            self.data = np.fromfile(path, dtype=dt)
        else:
            raise ValueError("FileType {} not supported".format(self.ftype))

    def load_data_free(self, path):
        data = []
        if self.ftype == FileTypes.ASCII:
            with open(path, "r") as ascii_file:
                ascii_file_reader = csv.reader(ascii_file, delimiter=";", quotechar='"')
                for row in ascii_file_reader:
                    f = float(row[0])
                    t = float(row[1])
                    n = row[2]
                    c = float(row[3])
                    data.append((f, t, n, c))

        elif self.ftype == FileTypes.BINARY:
            with open(path, "rb") as binary_file:
                counter = 0
                binary_file.seek(0)

                while counter < self.size:
                    # from (8byte float)
                    f = unpack("d", binary_file.read(8))[0]
                    ##to (8byte float)
                    t = unpack("d", binary_file.read(8))[0]
                    # length of label (4byte uint)
                    lol = unpack("i", binary_file.read(4))[0]
                    # the label (lol * byte)
                    n = binary_file.read(lol).decode("ISO-8859-1")
                    # confidence (4Byte float)
                    c = unpack("f", binary_file.read(4))[0]

                    data.append((f, t, n, c))
                    counter += 1
        else:
            raise ValueError("FileType {} not supported".format(self.ftype))

        self.data = np.array(data, LabelDataType.FREE.value)

    def load_data(self, path):
        if self.scheme.type == SchemeType.DISCRETE:
            self.load_data_discrete(path)
        elif self.scheme.type == SchemeType.CONTINUOUS:
            self.load_data_continuous(path)
        elif self.scheme.type == SchemeType.FREE:
            self.load_data_free(path)
        else:
            raise ValueError("SchemeType {} not supported".format(self.scheme.type))
        return self.data

    def load(self, path):
        self.load_header(path)
        self.load_data(path + "~")


if __name__ == "__main__":

    """discrete annotations"""
    anno_discrete_ascii = Anno()
    anno_discrete_ascii.load("Testfiles/discrete_ascii.annotation")

    anno_discrete_binary = Anno()
    anno_discrete_binary.load("Testfiles/discrete_binary.annotation")

    """continous annotations"""
    anno_cont_ascii = Anno()
    anno_cont_ascii.load("Testfiles/continuous_ascii.annotation")

    anno_cont_binary = Anno()
    anno_cont_binary.load("Testfiles/continuous_binary.annotation")

    """free annotations"""
    anno_free_ascii = Anno()
    anno_free_ascii.load("Testfiles/free_ascii.annotation")

    anno_free_binary = Anno()
    anno_free_binary.load("Testfiles/free_binary.annotation")

    ...



from abc import ABC, abstractmethod

from nova_utils.ssi_utils.ssi_anno_utils import Anno


REQUIREMENTS = []

class Processor(ABC):
    """
    Base class of a data processor. This interface builds the foundation for all data processing classes.
    """

    # List of dependencies that need to be installed when the script is loaded
    DEPENDENCIES = []

    # Flag to indicate whether the processed input belongs to one role or to multiple roles
    SINGLE_ROLE_INPUT = True

    def __init__(self, logger, request_form=None):
        self.logger = logger
        self.request_form = request_form
        self.model = None
        self.data = None
        self.output = None
        self.options = {}

        if self.request_form is not None:
            # Set Options
            logger.info("Setting options...")
            if request_form.get("optStr"):
                for k, v in dict(
                        option.split("=") for option in request_form["optStr"].split(";")
                ).items():
                    if v in ("True", "False"):
                        self.options[k] = True if v == "True" else False
                    elif v == "None":
                        self.options[k] = True if v == "True" else False
                    else:
                        self.options[k] = v
                    logger.info(k + "=" + v)
            logger.info("...done.")

    @abstractmethod
    def preprocess_sample(self, sample: dict ):
        """Preprocess data to convert between nova-server dataset iterator item to the raw model input as required in process_sample.

        Args:
            sample (dict):
        """
        return sample

    @abstractmethod
    def process_sample(self, sample):
        """Applying processing steps (e.g. feature extraction, data prediction etc... ) to the provided data."""
        return sample

    @abstractmethod
    def postprocess_sample(self, sample):
        """Apply any optional postprocessing to the data (e.g. scaling, mapping etc...)"""
        return sample

    def process_data(self, ds_iter) -> dict:
        """Returning a dictionary that contains the original keys from the dataset iterator and processed samples as value. Can be overwritten to customize the processing"""
        self.ds_iter = ds_iter
        processed = {k: [] for k in ds_iter.data_schemes.keys()}

        for sample in ds_iter:
            for k in ds_iter.data_schemes.keys():
                out = self.preprocess_sample(sample[k])
                out = self.process_sample(out)
                out = self.postprocess_sample(out)
                processed[k].append(out)

        return processed


class Trainer(Processor):
    """
    Base class of a Trainer. Implement this interface in your own class to build a model that is trainable from within nova
    """

    """Includes all the necessary files to run this script"""

    def __init__(self, logger, request_form=None):
        super().__init__(logger, request_form)

    @abstractmethod
    def train(self):
        """Trains a model with the given data."""
        raise NotImplemented

    @abstractmethod
    def save(self, path) -> str:
        """Stores the weights of the given model at the given path. Returns the path of the weights."""
        raise NotImplemented

    @abstractmethod
    def load(self, path):
        """Loads a model with the given path. Returns this model."""
        raise NotImplemented


class Predictor(Processor):
    """
    Base class of a data predictor. Implement this interface if you want to write annotations to a database
    """

    def __init__(self, logger, request_form=None):
        super().__init__(logger, request_form)

    @abstractmethod
    def to_anno(self, data) -> list[Anno]:
        """Converts the output of process_data to the correct annotation format to upload them to the database.
        !THE OUTPUT FORMAT OF THIS FUNCTION IS NOT YET FULLY DEFINED AND WILL CHANGE IN FUTURE RELEASES!

        Args:
            data (object): Data output of process_data function

        Returns:
            list: A list of annotation objects
        """
        raise NotImplemented


class Extractor(Processor):
    """
    Base class of a feature extractor. Implement this interface in your own class to build a feature extractor.
    """

    @property
    @abstractmethod
    def chainable(self):
        """Whether this extraction module can be followed by other extractors. If set to True 'to_ds_iterable()' must be implemented"""
        return False

    def __init__(self, logger, request_form=None):
        super().__init__(logger, request_form)

    @abstractmethod
    def to_stream(self, data: object) -> dict:
        """Converts the return value from process_data() to data stream chunk that can be processed by nova-server.

        Args:
            data (object): The data as returned by the process_data function of the Processor class


        Returns:
            dict: A dictionary mapping a stream identifier (usually composed using the role, signal, extracted feature name and sliding window parameters) to a tuple containing a chunk of the data as well as additional information.
        Each tuple has the form ( type (nova_types.DataTypes), sample_rate (double), data_chunk (numpy.ndarray) ). The shape of the data chunk should in the form of (n_frames, n_features)
        An arbitrary number of streams maybe returned.

        Example:
            ::

                {
                    'speaker_1.audio.mfcc[10ms,10ms,10ms]' : ( nova_utils.db_utils.nova_types.DataTypes.AUDIO, 100, [[0.0, 0.0, ... 0.0], [0.0, 0.0, ... 0.0] ... [0.0, 0.0, ... 0.0]] ),
                    'speaker_2.audio.mfcc[10ms,10ms,10ms]' : ( nova_utils.db_utils.nova_types.DataTypes.AUDIO, 100, [[0.0, 0.0, ... 0.0], [0.0, 0.0, ... 0.0] ... [0.0, 0.0, ... 0.0]] )
                }
        """
        raise NotImplemented



import xml.etree.ElementTree as ET
import numpy as np
from nova_utils.ssi_utils.ssi_data_types import FileTypes, string_to_enum
from nova_utils.ssi_utils.ssi_stream_utils import Stream
import os

class Sample:
    def __init__(self, streams = [], user_id = None, class_id = None, score = 0):
        self.streams = streams
        self.user_id = user_id
        self.class_id = class_id
        self.score = score

class SampleList:
    #def __init__(self, samples = None, size = 0,  missing = False, garbage=0, stream_paths = [], streams = [], classes = {}, user = {}, is_discrete = False):
    def __init__(self):
        self.samples = [] 
        self.samples = None
        self.size = 0
        self.missing = False
        self.garbage = 0
        self.stream_paths = []
        self.streams = []
        self.classes = {}
        self.user = {}
        self.is_discrete = False

    def load_header(self, path):
        tree = ET.parse(path)
        root = tree.getroot()

        for child in root:
            if child.tag ==  'info':
                for key,val in child.attrib.items():
                    if key == 'ftype':
                        self.ftype = string_to_enum(FileTypes, val)
                    elif key == 'size':
                        self.size = int(val)
                    elif key == 'missing':
                        self.missing = val
                    elif key == 'garbage':
                        self.garbage = int(val)
               
            elif child.tag == 'streams':
                for item in child:                    
                    for key,val in item.attrib.items():
                        if key == 'path':
                            self.stream_paths.append(val)              
            elif child.tag == 'classes':
                id = 0
                for item in child:
                    for key,val in item.attrib.items():
                        if key == "name":
                            class_name = val
                        elif key == "id":
                            id = val
                    self.classes[id] = class_name
                    id += 1
                if len(self.classes) > 0:
                    self.is_discrete = True
            elif child.tag == 'users':
                id = 0
                for item in child:
                    for key,val in item.attrib.items():
                        if key == "name":
                            user_name = val
                        elif key == "id":
                            id = val
                    self.user[id] = user_name
                    id += 1
        return self

    def load_samples(self, path):
        dt = {'names':('user_id', 'class_id', 'score', 'time'),
                          'formats':('i4', 'i4', 'f4', 'f8')}
        if self.ftype == FileTypes.ASCII:
            self.samples = np.loadtxt(path, dtype=dt,delimiter=" ")
        elif self.ftype == FileTypes.BINARY:
            self.samples = (np.fromfile(path, dtype=dt))
        else:
            raise ValueError('FileType {} not supported'.format(self))

    def load_streams(self, path):
        dir = os.path.dirname(path)
        
        for s in self.stream_paths:
            stream_path =  os.path.join(os.path. sep,os.getcwd(), dir, s + '.stream')
            stream = Stream().load(stream_path)
            self.streams.append(stream)

    def load(self, path):
        self.load_header(path)
        self.load_samples(path + '~')
        self.load_streams(path)
       
if __name__ == "__main__":
    sample_list_ascii = SampleList()
    sample_list_ascii.load("Testfiles/sl_ascii.samples")

    sample_list_binary = SampleList()
    sample_list_binary.load("Testfiles/sl_binary.samples")



from abc import ABC, abstractmethod


class DatasetIterable(ABC):
    """
    Abstract superclass for iterable Datasets, which can be fed into the Bridge classes for tensorflow and pytorch
    """

    def __init__(self, *args, split: str = "train", **kwargs):
        if split not in ["test", "train", "val"]:
            split = "train"
        self.split = split

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    @abstractmethod
    def __next__(self):
        raise NotImplementedError()

    @abstractmethod
    def get_output_info(self):
        raise NotImplementedError()



from enum import Enum
import numpy as np

class FileTypes(Enum):
        UNDEF = 0
        BINARY = 1
        ASCII = 2
        BIN_LZ4 = 3

class NPDataTypes(Enum):
    UNDEF = 0
    #CHAR = 1
    #UCHAR = 2
    SHORT = np.int16
    USHORT = np.uint16
    INT = np.int32
    UINT = np.uint32
    LONG = np.int64
    ULONG = np.uint64
    FLOAT = np.float32
    DOUBLE = np.float64
    LDOUBLE = np.float64
    #STRUCT = 12
    #IMAGE = 13
    BOOL = np.bool_

'''Helper'''
def string_to_enum(enum, string):
    for e in enum:
        if e.name == string:
            return e
    raise ValueError('{} not part of enumeration  {}'.format(string, enum))
